"""
container operations, including functional, deep dictionary syntactic sugar
"""

import itertools
from typing import Tuple, List, Dict, Iterable, Callable, TypeVar, Any, Union
from functools import reduce
from collections import OrderedDict

import numpy as np
import pandas as pd
from pandas.api.types import CategoricalDtype
import torch

from stefutil.prettier import pl


__all__ = [
    'get', 'set_', 'it_keys',
    'list_is_same_elms', 'chain_its', 'join_it', 'group_n', 'list_split', 'lst2uniq_ids', 'compress',
    'np_index', 'df_col2cat_col', 'pt_sample'
]


T = TypeVar('T')
K = TypeVar('K')


def get(dic: Dict, ks: str):
    """
    :param dic: Potentially multi-level dictionary
    :param ks: Potentially `.`-separated keys
    """
    ks = ks.split('.')
    _past_keys = []
    acc = dic
    for lvl, k in enumerate(ks):
        if k not in acc:
            _past_keys = pl.s('=>', c='m').join([pl.i(k) for k in _past_keys])
            d_log = {'past keys': _past_keys, 'available keys': list(acc.keys())}
            raise ValueError(f'{pl.i(k)} not found at level {pl.i(lvl+1)} with {pl.i(d_log)}')
        acc = acc[k]
        _past_keys.append(k)
    return acc


def set_(dic, ks, val):
    ks = ks.split('.')
    node = reduce(lambda acc, elm: acc[elm], ks[:-1], dic)
    node[ks[-1]] = val


def it_keys(dic, prefix=''):
    """
    :return: Generator for all potentially-nested keys
    """
    def _full(k_):
        return k_ if prefix == '' else f'{prefix}.{k_}'
    for k, v in dic.items():
        if isinstance(v, dict):
            for k__ in it_keys(v, prefix=_full(k)):
                yield k__
        else:
            yield _full(k)


def list_is_same_elms(lst: List[T]) -> bool:
    return all(l == lst[0] for l in lst)


def chain_its(its: Iterable[Iterable[T]]) -> Iterable[T]:
    """
    Chain multiple iterables
    """
    out = itertools.chain()
    for it in its:
        out = itertools.chain(out, it)
    return out


def join_it(it: Iterable[T], sep: T) -> Iterable[T]:
    """
    Generic join elements with separator element, like `str.join`
    """
    it = iter(it)

    curr = next(it, None)
    if curr is not None:
        yield curr
        curr = next(it, None)
    while curr is not None:
        yield sep
        yield curr
        curr = next(it, None)


def group_n(it: Iterable[T], n: int) -> Iterable[Tuple[T]]:
    """
    Slice iterable into groups of size n (last group included) by iteration order
    """
    # Credit: https://stackoverflow.com/a/8991553/10732321
    it = iter(it)
    while True:
        chunk = tuple(itertools.islice(it, n))
        if not chunk:
            return
        yield chunk


def list_split(lst: List[T], call: Callable[[T], bool]) -> List[List[T]]:
    """
    :return: Split a list by locations of elements satisfying a condition
    """
    return [list(g) for k, g in itertools.groupby(lst, call) if k]


def lst2uniq_ids(lst: List[T]) -> List[int]:
    """
    Each unique element in list assigned a unique id, in increasing order of iteration
    """
    elm2id = {v: k for k, v in enumerate(OrderedDict.fromkeys(lst))}
    return [elm2id[e] for e in lst]


def compress(lst: List[T]) -> List[Tuple[T, int]]:
    """
    :return: A compressed version of `lst`, as 2-tuple containing the occurrence counts
    """
    if not lst:
        return []
    return ([(lst[0], len(list(itertools.takewhile(lambda elm: elm == lst[0], lst))))]
            + compress(list(itertools.dropwhile(lambda elm: elm == lst[0], lst))))


def np_index(arr, idx):
    return np.where(arr == idx)[0][0]


def df_col2cat_col(df: pd.DataFrame, col_name: str, categories: List[str]) -> pd.DataFrame:
    """
    Enforced ordered categories to a column, the dataframe is modified in-place
    """
    cat = CategoricalDtype(categories=categories, ordered=True)  # Enforce order by definition
    df[col_name] = df[col_name].astype(cat, copy=False)
    return df


def pt_sample(d: Dict[K, Union[float, Any]]) -> K:
    """
    Sample a key from a dict based on confidence score as value
        Keys with confidence evaluated to false are ignored

    Internally uses `torch.multinomial`
    """
    d_keys = {k: v for k, v in d.items() if v}  # filter out `None`s
    keys, weights = zip(*d_keys.items())
    return keys[torch.multinomial(torch.tensor(weights), 1, replacement=True).item()]


if __name__ == '__main__':
    from icecream import ic

    def check_get():
        d = {'a': {'b': {'c': 1, 'd': 2}, 'e': 3}, 'f': 4}
        ic(d)
        ic(get(d, 'a.b.c'))
        ic(get(d, 'a.b.e'))
    check_get()



"""
plotting

see also `StefUtil.save_fig`
"""

from typing import List, Iterable, Callable, Any, Union

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

from stefutil.prettier import ca
from stefutil.container import df_col2cat_col


__all__ = [
    'LN_KWARGS',
    'change_bar_width', 'vals2colors', 'set_color_bar', 'barplot',
]


plt.rc('figure', figsize=(16, 9))
plt.rc('figure.constrained_layout', use=True)
plt.rc('text.latex', preamble='\n'.join([
    r'\usepackage{nicefrac}',
    r'\usepackage{helvet}',
    r'\usepackage{sansmath}',  # render math sans-serif
    r'\sansmath'
]))
sns.set_style('darkgrid')
sns.set_context(rc={'grid.linewidth': 0.5})


LN_KWARGS = dict(marker='o', ms=0.3, lw=0.25)  # matplotlib line plot default args


def change_bar_width(ax, width: float = 0.5, orient: str = 'v'):
    """
    Modifies the bar width of a matplotlib bar plot

    Credit: https://stackoverflow.com/a/44542112/10732321
    """
    ca(bar_orient=orient)
    is_vert = orient in ['v', 'vertical']
    for patch in ax.patches:
        current_width = patch.get_width() if is_vert else patch.get_height()
        diff = current_width - width
        patch.set_width(width) if is_vert else patch.set_height(width)
        patch.set_x(patch.get_x() + diff * 0.5) if is_vert else patch.set_y(patch.get_y() + diff * 0.5)


def vals2colors(
        vals: Iterable[float], color_palette: str = 'Spectral_r', gap: float = None,
) -> List:
    """
    Map an iterable of values to corresponding colors given a color map

    :param vals: Values to map color
    :param color_palette: seaborn color map
    :param gap: Ratio of difference between min and max values
        If given, reduce visual spread/difference of colors
        Intended for a less drastic color at the extremes
    """
    vals = np.asarray(vals)
    cmap = sns.color_palette(color_palette, as_cmap=True)
    mi, ma = np.min(vals), np.max(vals)
    if gap is not None:
        diff = ma - mi
        mi, ma = mi - diff * gap, ma + diff * gap
    norm = (vals - mi) / (ma - mi)
    return cmap(norm)


def set_color_bar(vals, ax, color_palette: str = 'Spectral_r', orientation: str = 'vertical'):
    """
    Set give axis to show the color bard
    """
    vals = np.asarray(vals)
    norm = plt.Normalize(vmin=np.min(vals), vmax=np.max(vals))
    sm = plt.cm.ScalarMappable(cmap=color_palette, norm=norm)
    sm.set_array([])
    plt.sca(ax)
    plt.grid(False)
    plt.colorbar(sm, cax=ax, orientation=orientation)
    # plt.xlabel('colorbar')  # doesn't seem to work


def barplot(
        data: pd.DataFrame = None,
        x: Union[Iterable, str] = None, y: Union[Iterable[float], str] = None,
        x_order: Iterable[str] = None,
        orient: str = 'v', with_value: bool = False, width: [float, bool] = 0.5,
        xlabel: str = None, ylabel: str = None, yscale: str = None, title: str = None,
        ax=None, palette: Union[str, List, Any] = 'husl', callback: Callable[[plt.Axes], None] = None,
        show: bool = True,
        **kwargs
):
    ca(bar_orient=orient)
    if data is not None:
        df = data
        assert isinstance(x, str) and isinstance(y, str)
        df['x'], df['y'] = df[x], df[y]
    else:
        df = pd.DataFrame([dict(x=x_, y=y_) for x_, y_ in zip(x, y)])
        x_order = x
    if x_order is not None:
        x_order: List[str]
        df_col2cat_col(df, 'x', categories=x_order)
    is_vert = orient in ['v', 'vertical']
    x, y = ('x', 'y') if is_vert else ('y', 'x')
    if ax:
        kwargs['ax'] = ax
    if palette is not None:
        kwargs['palette'] = palette
    ax = sns.barplot(data=df, x=x, y=y, **kwargs)
    if with_value:
        ax.bar_label(ax.containers[0])
    if width:
        change_bar_width(ax, width, orient=orient)
    ax.set_xlabel(xlabel) if is_vert else ax.set_ylabel(xlabel)  # if None just clears the label
    ax.set_ylabel(ylabel) if is_vert else ax.set_xlabel(ylabel)
    if yscale:
        ax.set_yscale(yscale) if is_vert else ax.set_xscale(yscale)
    if title:
        ax.set_title(title)
    if callback:
        callback(ax)
    if show:
        plt.show()
    return ax



"""
concurrency

intended for (potentially heavy) data processing
"""

import os
import concurrent.futures
from typing import List, Tuple, Dict, Iterable, Callable, TypeVar, Union

import numpy as np
from tqdm.std import tqdm as std_tqdm  # root for type check
from tqdm.auto import tqdm
from tqdm.contrib import concurrent as tqdm_concurrent

from stefutil.prettier import ca


__all__ = ['conc_map', 'batched_conc_map']


T = TypeVar('T')
K = TypeVar('K')

MapFn = Callable[[T], K]
BatchedMapFn = Callable[[Tuple[List[T], int, int]], List[K]]


def conc_map(
        fn: MapFn, it: Iterable[T], with_tqdm: Union[bool, Dict] = False, n_worker: int = os.cpu_count(), mode: str = 'thread'
) -> Iterable[K]:
    """
    Wrapper for `concurrent.futures.map`

    :param fn: A function
    :param it: A list of elements
    :param with_tqdm: If true, progress bar is shown
        If dict, treated as `tqdm` concurrent kwargs  # FYT `chunksize` is helpful
    :param n_worker: Number of concurrent workers
    :param mode: One of ['thread', 'process']
        Function has to be pickleable if 'process'
    :return: Iterator of `lst` elements mapped by `fn` with thread concurrency
    """
    ca.check_mismatch('Concurrency Mode', mode, ['thread', 'process'])
    if with_tqdm:
        cls = tqdm_concurrent.thread_map if mode == 'thread' else tqdm_concurrent.process_map
        args = (isinstance(with_tqdm, dict) and with_tqdm) or dict()
        return cls(fn, it, max_workers=n_worker, **args)
    else:
        cls = concurrent.futures.ThreadPoolExecutor if mode == 'thread' else concurrent.futures.ProcessPoolExecutor
        with cls(max_workers=n_worker) as executor:
            return executor.map(fn, it)


"""
classes instead of nested functions, pickleable for multiprocessing
"""


class Map:
    def __init__(self, fn, pbar=None):
        self.fn = fn
        self.pbar = pbar

    def __call__(self, x):
        ret = self.fn(x)
        if self.pbar:
            self.pbar.update(1)
        return ret


class BatchedMap:
    def __init__(self, fn, is_batched_fn: bool, pbar=None):
        self.fn = fn if is_batched_fn else Map(fn, pbar)
        self.pbar = pbar
        self.is_batched_fn = is_batched_fn

    def __call__(self, args):
        # adhere to single-argument signature for `conc_map`
        lst, s, e = args
        if self.is_batched_fn:
            ret = self.fn(lst[s:e])
            if self.pbar:
                # TODO: update on the element level may not give a good estimate of completion if too large a batch
                self.pbar.update(e-s + 1)
            return ret
        else:
            return [self.fn(lst[i]) for i in range(s, e)]


def batched_conc_map(
        fn: Union[MapFn, BatchedMapFn],
        it: Union[Iterable[T], List[T], np.array], n: int = None, n_worker: int = os.cpu_count(),
        batch_size: int = None,
        with_tqdm: Union[bool, dict, tqdm] = False,
        is_batched_fn: bool = False,
        mode: str = 'thread'
) -> List[K]:
    """
    Batched concurrent mapping, map elements in list in batches
    Operates on batch/subset of `lst` elements given inclusive begin & exclusive end indices

    :param fn: A map function that operates on a single element
    :param it: A list of elements to map
    :param n: #elements to map if `it` is not Sized
    :param n_worker: Number of concurrent workers
    :param batch_size: Number of elements for each sub-process worker
        Inferred based on number of workers if not given
    :param with_tqdm: If true, progress bar is shown
        progress is shown on an element-level if possible
    :param is_batched_fn: If true, `conc_map` is called on the function passed in,
        otherwise, A batched version is created internally
    :param mode: One of ['thread', 'process']

    .. note:: Concurrently is not invoked if too little list elements given number of workers
        Force concurrency with `batch_size`
    """
    n = n or len(it)
    if (n_worker > 1 and n > n_worker * 4) or batch_size:  # factor of 4 is arbitrary, otherwise not worse the overhead
        preprocess_batch = batch_size or round(n / n_worker / 2)
        strts: List[int] = list(range(0, n, preprocess_batch))
        ends: List[int] = strts[1:] + [n]  # inclusive begin, exclusive end
        lst_out = []

        pbar = None
        if with_tqdm:
            tqdm_args = dict(mode=mode, n_worker=n_worker)
            if mode == 'thread':  # Able to show progress on element level
                # so create such a progress bar & disable for `conc_map`
                tqdm_args['with_tqdm'] = False
                if isinstance(with_tqdm, bool):
                    pbar = tqdm(total=n)
                elif isinstance(with_tqdm, dict):
                    _args = dict(total=n)
                    _args.update(with_tqdm)
                    pbar = tqdm(**_args)
                else:
                    assert isinstance(with_tqdm, std_tqdm)
                    pbar = with_tqdm
            else:  # `process`, have to rely on `tqdm.concurrent` which shows progress on batch level, see `conc_map`
                tqdm_args['with_tqdm'] = with_tqdm
        else:
            tqdm_args = dict(with_tqdm=False)

        batched_map = BatchedMap(fn, is_batched_fn, pbar)
        map_out = conc_map(fn=batched_map, it=[(it, s, e) for s, e in zip(strts, ends)], **tqdm_args)
        for lst_ in map_out:
            lst_out.extend(lst_)
        return lst_out
    else:
        gen = tqdm(it) if with_tqdm else it
        if is_batched_fn:
            args = gen, 0, n
            return fn(*args)
        else:
            return [fn(x) for x in gen]



"""
os-related
"""

import os
import pathlib
from typing import Union


__all__ = ['get_hostname', 'stem']


def get_hostname() -> str:
    return os.uname().nodename


def stem(path: Union[str, pathlib.Path], keep_ext=False) -> str:
    """
    :param path: A potentially full path to a file
    :param keep_ext: If True, file extensions is preserved
    :return: The file name, without parent directories
    """
    return os.path.basename(path) if keep_ext else pathlib.Path(path).stem



"""
machine learning
"""

import sys
from typing import Union

import torch

from stefutil.prettier import fmt_num

__all__ = ['model_param_size', 'get_model_num_trainable_parameter', 'is_on_colab']


def get_torch_device():
    return 'cuda' if torch.cuda.is_available() else 'cpu'


def model_param_size(m: torch.nn.Module, as_str=True) -> Union[int, str]:
    num = sum(p.numel() for p in m.parameters())
    return fmt_num(num) if as_str else num


def get_model_num_trainable_parameter(model: torch.nn.Module, readable: bool = True) -> Union[int, str]:
    n = sum(p.numel() for p in model.parameters() if p.requires_grad)
    return fmt_num(n) if readable else n


def is_on_colab() -> bool:
    return 'google.colab' in sys.modules



"""
enhanced-built-in function
"""

__all__ = ['vars_']


def vars_(obj, include_private=False):
    """
    :return: A variant of `vars` that returns all properties and corresponding values in `dir`, except the
    generic ones that begins with `_`
    """
    def is_relevant():
        if include_private:
            return lambda a: not a.startswith('__')
        else:
            return lambda a: not a.startswith('__') and not a.startswith('_')
    attrs = filter(is_relevant(), dir(obj))
    return {a: getattr(obj, a) for a in attrs}



"""
function-level manipulation
"""

import sys
import time
from typing import Callable, Union


__all__ = ['profile_runtime', 'RecurseLimit']


def profile_runtime(callback: Callable, sleep: Union[float, int] = None):
    import cProfile
    import pstats
    profiler = cProfile.Profile()
    profiler.enable()
    callback()
    profiler.disable()
    stats = pstats.Stats(profiler).sort_stats('cumtime')
    if sleep:    # Sometimes, the top rows in `print_states` are now shown properly
        time.sleep(sleep)
    stats.print_stats()


class RecurseLimit:
    # credit: https://stackoverflow.com/a/50120316/10732321
    def __init__(self, limit):
        self.limit = limit

    def __enter__(self):
        self.old_limit = sys.getrecursionlimit()
        sys.setrecursionlimit(self.limit)

    def __exit__(self, kind, value, tb):
        sys.setrecursionlimit(self.old_limit)



"""
Project & project file structure related
"""

import os
import json
from os.path import join as os_join
from typing import Dict

import matplotlib.pyplot as plt

from stefutil.container import get
from stefutil.prettier import now


__all__ = ['StefConfig', 'StefUtil']


class StefConfig:
    """
    the one-stop place for package-level constants, expects a json file
    """
    def __init__(self, config_file: str):
        self.config_file = config_file
        with open(config_file, 'r') as f:
            self.d = json.load(f)

    def __call__(self, keys: str):
        """
        Retrieves the queried attribute value from the config file

        Loads the config file on first call.
        """
        return get(self.d, keys)


class StefUtil:
    """
    Effectively curried functions with my enforced project & dataset structure
        Pass in file paths
    """
    plot_dir = 'plot'
    eval_dir = 'eval'

    def __init__(
            self, base_path: str = None, project_dir: str = None, package_name: str = None,
            dataset_dir: str = None, model_dir: str = None
    ):
        """
        :param base_path: Absolute system path for root directory that contains a project folder & a data folder
        :param project_dir: Project root directory name that contains a folder for main source files
        :param package_name: python package/Module name which contain main source files
        :param dataset_dir: Directory name that contains datasets
        :param model_dir: Directory name that contains trained models
        """
        self.base_path = base_path
        self.proj_dir = project_dir
        self.pkg_nm = package_name
        self.dset_dir = dataset_dir
        self.model_dir = model_dir

        self.proj_path = os_join(self.base_path, self.proj_dir)
        self.dset_path = os_join(self.base_path, self.dset_dir)
        self.model_path = os_join(self.base_path, self.model_dir)
        self.plot_path = os_join(self.base_path, self.proj_dir, StefUtil.plot_dir)
        self.eval_path = os_join(self.base_path, self.proj_dir, StefUtil.eval_dir)
        os.makedirs(self.plot_path, exist_ok=True)
        os.makedirs(self.eval_path, exist_ok=True)

    def save_fig(
            self, title, save=True, prefix_time: bool = True, save_path: str = None, time_args: Dict = None
    ):
        """
        :param title: Rendered figure title
        :param save: If true, figure is saved to project plot directory
            No effect otherwise
        :param prefix_time: If true, timestamp is prefixed before filename
            Otherwise, timestamp is appended to the end
        :param save_path: disk path to save figure
        :param time_args: `now` arguments
        """
        if save:
            args = dict(fmt='short-date') | (time_args or dict())
            t = now(**args)
            if prefix_time:
                fnm = f'{t}_{title}.png'
            else:
                fnm = f'{title}, {t}.png'
            plt.savefig(os_join((save_path or self.plot_path), fnm), dpi=300)



"""
prettier & prettier logging
"""

import os
import re
import sys
import json
import math
import pprint
import logging
import datetime
from typing import Tuple, List, Dict, Iterable, Any, Union
from pygments import highlight, lexers, formatters
from collections import OrderedDict
from collections.abc import Sized

import pandas as pd
from torch.utils.tensorboard import SummaryWriter
from transformers import Trainer, TrainerCallback
import sty
import colorama
from tqdm.auto import tqdm
from icecream import IceCreamDebugger

from stefutil.primitive import is_float


__all__ = [
    'fmt_num', 'fmt_sizeof', 'fmt_delta', 'sec2mmss', 'round_up_1digit', 'nth_sig_digit',
    'MyIceCreamDebugger', 'mic',
    'PrettyLogger', 'pl',
    'hex2rgb', 'MyTheme', 'MyFormatter', 'get_logger',
    'CheckArg', 'ca',
    'now',
    'MlPrettier', 'MyProgressCallback', 'LogStep'
]


pd.set_option('expand_frame_repr', False)
pd.set_option('display.precision', 2)
pd.set_option('max_colwidth', 40)
pd.set_option('display.max_columns', None)
pd.set_option('display.min_rows', 16)


def fmt_num(num: Union[float, int], suffix: str = '') -> str:
    """
    Convert number to human-readable format, in e.g. Thousands, Millions
    """
    for unit in ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z']:
        if abs(num) < 1000.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1000.0
    return "%.1f%s%s" % (num, 'Y', suffix)


def fmt_sizeof(num: int, suffix='B') -> str:
    """ Converts byte size to human-readable format """
    for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)


def fmt_delta(secs: Union[int, float, datetime.timedelta]) -> str:
    if isinstance(secs, datetime.timedelta):
        secs = 86400 * secs.days + secs.seconds + (secs.microseconds/1e6)
    if secs >= 86400:
        d = secs // 86400  # // floor division
        return f'{round(d)}d{fmt_delta(secs - d * 86400)}'
    elif secs >= 3600:
        h = secs // 3600
        return f'{round(h)}h{fmt_delta(secs - h * 3600)}'
    elif secs >= 60:
        m = secs // 60
        return f'{round(m)}m{fmt_delta(secs - m * 60)}'
    else:
        return f'{round(secs)}s'


def sec2mmss(sec: int) -> str:
    return str(datetime.timedelta(seconds=sec))[2:]


def round_up_1digit(num: int):
    d = math.floor(math.log10(num))
    fact = 10**d
    return math.ceil(num/fact) * fact


def nth_sig_digit(flt: float, n: int = 1) -> float:
    """
    :return: first n-th significant digit of `sig_d`
    """
    return float('{:.{p}g}'.format(flt, p=n))


class MyIceCreamDebugger(IceCreamDebugger):
    def __init__(self, output_width: int = 120, **kwargs):
        self._output_width = output_width
        kwargs.update(argToStringFunction=lambda x: pprint.pformat(x, width=output_width))
        super().__init__(**kwargs)
        self.lineWrapWidth = output_width

    @property
    def output_width(self):
        return self._output_width

    @output_width.setter
    def output_width(self, value):
        if value != self._output_width:
            self._output_width = value
            self.lineWrapgitWidth = value
            self.argToStringFunction = lambda x: pprint.pformat(x, width=value)


mic = MyIceCreamDebugger()


class PrettyLogger:
    """
    My logging w/ color & formatting, and a lot of syntactic sugar
    """
    reset = colorama.Fore.RESET + colorama.Back.RESET + colorama.Style.RESET_ALL
    key2c = dict(
        log='',
        warn=colorama.Fore.YELLOW,
        error=colorama.Fore.RED,
        err=colorama.Fore.RED,
        success=colorama.Fore.GREEN,
        suc=colorama.Fore.GREEN,
        info=colorama.Fore.BLUE,
        i=colorama.Fore.BLUE,
        w=colorama.Fore.RED,

        y=colorama.Fore.YELLOW,
        yellow=colorama.Fore.YELLOW,
        red=colorama.Fore.RED,
        r=colorama.Fore.RED,
        green=colorama.Fore.GREEN,
        g=colorama.Fore.GREEN,
        blue=colorama.Fore.BLUE,
        b=colorama.Fore.BLUE,

        m=colorama.Fore.MAGENTA
    )

    @staticmethod
    def log(s, c: str = 'log', c_time='green', as_str=False, bold: bool = False, pad: int = None):
        """
        Prints `s` to console with color `c`
        """
        need_reset = False
        if c in PrettyLogger.key2c:
            c = PrettyLogger.key2c[c]
            need_reset = True
        if bold:
            c += colorama.Style.BRIGHT
            need_reset = True
        reset = PrettyLogger.reset if need_reset else ''
        if as_str:
            return f'{c}{s:>{pad}}{reset}' if pad is not None else f'{c}{s}{reset}'
        else:
            print(f'{c}{PrettyLogger.log(now(), c=c_time, as_str=True)}| {s}{reset}')

    @staticmethod
    def s(s, c: str = None, bold: bool = False, with_color: bool = True) -> str:
        """
        syntactic sugar for return string instead of print
        """
        c = c if with_color else ''  # keeping the same signature with logging specific types for `lognc`
        return PrettyLogger.log(s, c=c, as_str=True, bold=bold)

    @staticmethod
    def i(s, **kwargs):
        """
        Syntactic sugar for logging `info` as string
        """
        if isinstance(s, dict):
            return PrettyLogger._dict(s, **kwargs)
        elif isinstance(s, list):
            return PrettyLogger._list(s, **kwargs)
        elif isinstance(s, tuple):
            return PrettyLogger._tuple(s, **kwargs)
        else:
            kwargs_ = dict(c='i')
            kwargs_.update(kwargs)
            return PrettyLogger.s(s, **kwargs_)

    @staticmethod
    def pa(s):
        assert isinstance(s, dict)
        return PrettyLogger.i(s, for_path=True, with_color=False)

    @staticmethod
    def nc(s, **kwargs):
        """
        Syntactic sugar for `i` w/o color
        """
        return PrettyLogger.i(s, with_color=False, **kwargs)

    @staticmethod
    def id(d: Dict) -> str:
        """
        Indented
        """
        return json.dumps(d, indent=4)

    @staticmethod
    def fmt(s) -> str:
        """
        colored by `pygments` & with indent
        """
        return highlight(PrettyLogger.id(s), lexers.JsonLexer(), formatters.TerminalFormatter())

    @staticmethod
    def _iter(it: Iterable, with_color=True, pref: str = '[', post: str = ']'):
        if with_color:
            pref, post = PrettyLogger.s(pref, c='m'), PrettyLogger.s(post, c='m')
        lst = [PrettyLogger.i(e, with_color=with_color) for e in it]
        return f'{pref}{", ".join(lst)}{post}'

    @staticmethod
    def _list(lst: List, with_color=True):
        return PrettyLogger._iter(lst, with_color=with_color, pref='[', post=']')

    @staticmethod
    def _tuple(tpl: Tuple, with_color=True):
        return PrettyLogger._iter(tpl, with_color=with_color, pref='(', post=')')

    @staticmethod
    def _dict(d: Dict = None, with_color=True, pad_float: int = 5, sep=': ', for_path: bool = False, **kwargs) -> str:
        """
        Syntactic sugar for logging dict with coloring for console output
        """
        def _log_val(v):
            if isinstance(v, (dict, list, tuple)):
                return PrettyLogger.i(v, with_color=with_color)
            else:
                if is_float(v):  # Pad only normal, expected floats, intended for metric logging
                    if is_float(v, no_int=True, no_sci=True):
                        v = float(v)
                        if with_color:
                            return PrettyLogger.log(v, c='i', as_str=True, pad=pad_float)
                        else:
                            return f'{v:>{pad_float}}'
                    else:
                        return PrettyLogger.i(v) if with_color else v
                else:
                    return PrettyLogger.i(v) if with_color else v
        d = d or kwargs or dict()
        if for_path:
            assert not with_color  # sanity check
            sep = '='
        if with_color:
            sep = PrettyLogger.s(sep, c='m')
        pairs = (f'{k}{sep}{_log_val(v)}' for k, v in d.items())
        pref, post = '{', '}'
        if with_color:
            pref, post = PrettyLogger.s(pref, c='m'), PrettyLogger.s(post, c='m')
        return pref + ', '.join(pairs) + post


pl = PrettyLogger()


def hex2rgb(hx: str, normalize=False) -> Union[Tuple[int], Tuple[float]]:
    # Modified from https://stackoverflow.com/a/62083599/10732321
    if not hasattr(hex2rgb, 'regex'):
        hex2rgb.regex = re.compile(r'#[a-fA-F\d]{3}(?:[a-fA-F\d]{3})?$')
    m = hex2rgb.regex.match(hx)
    assert m is not None
    if len(hx) <= 4:
        ret = tuple(int(hx[i]*2, 16) for i in range(1, 4))
    else:
        ret = tuple(int(hx[i:i+2], 16) for i in range(1, 7, 2))
    return tuple(i/255 for i in ret) if normalize else ret


class MyTheme:
    """
    Theme based on `sty` and `Atom OneDark`
    """
    COLORS = OrderedDict([
        ('yellow', 'E5C07B'),
        ('green', '00BA8E'),
        ('blue', '61AFEF'),
        ('cyan', '2AA198'),
        ('red', 'E06C75'),
        ('purple', 'C678DD')
    ])
    yellow, green, blue, cyan, red, purple = (
        hex2rgb(f'#{h}') for h in ['E5C07B', '00BA8E', '61AFEF', '2AA198', 'E06C75', 'C678DD']
    )

    @staticmethod
    def set_color_type(t: str):
        """
        Sets the class attribute accordingly

        :param t: One of ['rgb`, `sty`]
            If `rgb`: 3-tuple of rgb values
            If `sty`: String for terminal styling prefix
        """
        for color, hex_ in MyTheme.COLORS.items():
            val = hex2rgb(f'#{hex_}')  # For `rgb`
            if t == 'sty':
                setattr(sty.fg, color, sty.Style(sty.RgbFg(*val)))
                val = getattr(sty.fg, color)
            setattr(MyTheme, color, val)


class MyFormatter(logging.Formatter):
    """
    Modified from https://stackoverflow.com/a/56944256/10732321

    Default styling: Time in green, metadata indicates severity, plain log message
    """
    RESET = sty.rs.fg + sty.rs.bg + sty.rs.ef

    MyTheme.set_color_type('sty')
    yellow, green, blue, cyan, red, purple = (
        MyTheme.yellow, MyTheme.green, MyTheme.blue, MyTheme.cyan, MyTheme.red, MyTheme.purple
    )

    KW_TIME = '%(asctime)s'
    KW_MSG = '%(message)s'
    KW_LINENO = '%(lineno)d'
    KW_FNM = '%(filename)s'
    KW_FUNC_NM = '%(funcName)s'
    KW_NAME = '%(name)s'

    DEBUG = INFO = BASE = RESET
    WARN, ERR, CRIT = yellow, red, purple
    CRIT += sty.Style(sty.ef.bold)

    LVL_MAP = {  # level => (abbreviation, style)
        logging.DEBUG: ('DBG', DEBUG),
        logging.INFO: ('INFO', INFO),
        logging.WARNING: ('WARN', WARN),
        logging.ERROR: ('ERR', ERR),
        logging.CRITICAL: ('CRIT', CRIT)
    }

    def __init__(self, with_color=True, color_time=green):
        super().__init__()
        self.with_color = with_color

        sty_kw, reset = MyFormatter.blue, MyFormatter.RESET
        color_time = f'{color_time}{MyFormatter.KW_TIME}{sty_kw}|{reset}'

        def args2fmt(args_):
            if self.with_color:
                return color_time + self.fmt_meta(*args_) + f'{sty_kw}: {reset}{MyFormatter.KW_MSG}' + reset
            else:
                return f'{MyFormatter.KW_TIME}| {self.fmt_meta(*args_)}: {MyFormatter.KW_MSG}'

        self.formats = {level: args2fmt(args) for level, args in MyFormatter.LVL_MAP.items()}
        self.formatter = {
            lv: logging.Formatter(fmt, datefmt='%Y-%m-%d %H:%M:%S') for lv, fmt in self.formats.items()
        }

    def fmt_meta(self, meta_abv, meta_style=None):
        if self.with_color:
            return f'{MyFormatter.purple}[{MyFormatter.KW_NAME}]' \
               f'{MyFormatter.blue}::{MyFormatter.purple}{MyFormatter.KW_FUNC_NM}' \
               f'{MyFormatter.blue}::{MyFormatter.purple}{MyFormatter.KW_FNM}' \
               f'{MyFormatter.blue}:{MyFormatter.purple}{MyFormatter.KW_LINENO}' \
               f'{MyFormatter.blue}:{meta_style}{meta_abv}{MyFormatter.RESET}'
        else:
            return f'[{MyFormatter.KW_NAME}] {MyFormatter.KW_FUNC_NM}::{MyFormatter.KW_FNM}' \
                   f':{MyFormatter.KW_LINENO}, {meta_abv}'

    def format(self, entry):
        return self.formatter[entry.levelno].format(entry)


def get_logger(name: str, typ: str = 'stdout', file_path: str = None) -> logging.Logger:
    """
    :param name: Name of the logger
    :param typ: Logger type, one of [`stdout`, `file-write`]
    :param file_path: File path for file-write logging
    """
    assert typ in ['stdout', 'file-write']
    logger = logging.getLogger(f'{name} file write' if typ == 'file-write' else name)
    logger.handlers = []  # A crude way to remove prior handlers, ensure only 1 handler per logger
    logger.setLevel(logging.DEBUG)
    if typ == 'stdout':
        handler = logging.StreamHandler(stream=sys.stdout)  # stdout for my own coloring
    else:  # `file-write`
        if not file_path:
            raise ValueError(f'{pl.i(file_path)} must be specified for {pl.i("file-write")} logging')
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        handler = logging.FileHandler(file_path)
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(MyFormatter(with_color=typ == 'stdout'))
    logger.addHandler(handler)
    logger.propagate = False
    return logger


class MlPrettier:
    """
    My utilities for deep learning training logging
    """
    no_prefix = ('epoch', 'step')

    def __init__(
            self, ref: Dict[str, Any] = None, metric_keys: List[str] = None, no_prefix: Iterable[str] = no_prefix,
            with_color: bool = False
    ):
        """
        :param ref: Reference that are potentially needed
            i.e. for logging epoch/step, need the total #
        :param metric_keys: keys that are considered metric
            Will be logged in [0, 100]
        """
        self.ref = ref
        self.metric_keys = metric_keys or ['acc', 'recall', 'auc']
        self.no_prefix = no_prefix
        self.with_color = with_color

    def __call__(self, d: Union[str, Dict], val=None) -> Union[Any, Dict[str, Any]]:
        """
        :param d: If str, prettify a single value
            Otherwise, prettify a dict
        """
        is_dict = isinstance(d, dict)
        if not ((isinstance(d, str) and val is not None) or is_dict):
            raise ValueError('Either a key-value pair or a mapping is expected')
        if is_dict:
            d: Dict
            return {k: self.single(k, v) for k, v in d.items()}
        else:
            return self.single(d, val)

    def single(self, key: str = None, val: Any = None) -> Union[str, List[str], Dict[str, Any]]:
        """
        `val` processing is infered based on key
        """
        if key in ['step', 'epoch']:
            k = next(iter(k for k in self.ref.keys() if key in k))
            lim = self.ref[k]
            assert isinstance(val, (int, float))
            len_lim = len(str(lim))
            if isinstance(val, int):
                s_val = f'{val:>{len_lim}}'
            else:
                fmt = f'%{len_lim + 4}.3f'
                s_val = fmt % val
            if self.with_color:
                return f'{pl.i(s_val)}/{pl.i(lim)}'
            else:
                return f'{s_val}/{lim}'  # Pad integer
        elif 'loss' in key:
            return f'{round(val, 4):7.4f}'
        elif any(k in key for k in self.metric_keys):  # custom in-key-ratio metric
            def _single(v):
                return f'{round(v * 100, 2):6.2f}' if v is not None else '-'

            if isinstance(val, list):
                return [_single(v) for v in val]
            elif isinstance(val, dict):
                return {k: _single(v) for k, v in val.items()}
            else:
                return _single(val)
        elif 'learning_rate' in key or 'lr' in key:
            return f'{round(val, 7):.3e}'
        else:
            return val

    def should_add_split_prefix(self, key: str) -> bool:
        """
        Whether to add split prefix to the key
        """
        return key not in self.no_prefix

    def add_split_prefix(self, d: Dict[str, Any], split: str = None):
        if split is None:
            return d
        else:
            return {f'{split}/{k}' if self.should_add_split_prefix(k) else k: v for k, v in d.items()}


class MyProgressCallback(TrainerCallback):
    """
    My modification to the HF progress callback

    1. Effectively remove all logging, keep only the progress bar w.r.t. this callback
    2. Train tqdm for each epoch only
    3. Option to disable progress bar for evaluation

    Expects to start from whole epochs
    """
    def __init__(self, train_only: bool = False):
        """
        :param train_only: If true, disable progress bar for evaluation
        """
        self.training_bar = None
        self.prediction_bar = None

        self.train_only = train_only
        self.step_per_epoch = None
        self.current_step = None

    @staticmethod
    def _get_steps_per_epoch(state):
        assert state.max_steps % state.num_train_epochs == 0
        return state.max_steps // state.num_train_epochs

    @staticmethod
    def _get_curr_epoch(state, is_eval: bool = False) -> str:
        n_ep = int(state.epoch)
        if not is_eval:  # heuristic judging by the eval #epoch shown
            n_ep += 1

        return MlPrettier(ref=dict(epoch=state.num_train_epochs), with_color=True)('epoch', n_ep)

    def on_epoch_begin(self, args, state, control, **kwargs):
        if state.is_local_process_zero:
            if not self.step_per_epoch:
                self.step_per_epoch = MyProgressCallback._get_steps_per_epoch(state)
            ep = MyProgressCallback._get_curr_epoch(state)
            self.training_bar = tqdm(total=self.step_per_epoch, desc=f'Train Epoch {ep}', unit='ba')
        self.current_step = 0

    def on_train_begin(self, args, state, control, **kwargs):
        pass

    def on_epoch_end(self, args, state, control, **kwargs):
        if state.is_local_process_zero:
            self.training_bar.close()
            self.training_bar = None

    def on_step_end(self, args, state, control, **kwargs):
        if state.is_local_process_zero:
            self.training_bar.update(1)

    def on_prediction_step(self, args, state, control, eval_dataloader=None, **kwargs):
        if not self.train_only:
            if state.is_local_process_zero and isinstance(eval_dataloader.dataset, Sized):
                if self.prediction_bar is None:
                    ep = MyProgressCallback._get_curr_epoch(state, is_eval=True)
                    desc = f'Eval Epoch {ep}'
                    self.prediction_bar = tqdm(
                        desc=desc, total=len(eval_dataloader), leave=self.training_bar is None, unit='ba'
                    )
                self.prediction_bar.update(1)

    def on_evaluate(self, args, state, control, **kwargs):
        if not self.train_only:
            if state.is_local_process_zero:
                if self.prediction_bar is not None:
                    self.prediction_bar.close()
                self.prediction_bar = None

    def on_log(self, args, state, control, logs=None, **kwargs):
        if state.is_local_process_zero and self.training_bar is not None:
            _ = logs.pop("total_flos", None)

    def on_train_end(self, args, state, control, **kwargs):
        pass

    @staticmethod
    def get_current_progress_bar(trainer: Trainer):
        """
        Intended for adding per-step metrics to the progress bar during HF training

        This is a hack,
            since HF API don't support per-step callback not to mention exposing those metrics to the progress bar
        """
        callback = next(cb for cb in trainer.callback_handler.callbacks if isinstance(cb, MyProgressCallback))
        return callback.training_bar if trainer.model.training else callback.prediction_bar


class LogStep:
    """
    My typical terminal, file-write, tqdm logging for a single step
    """
    def __init__(
            self, trainer: Trainer = None, prettier: MlPrettier = None,
            logger: logging.Logger = None, file_logger: logging.Logger = None,
            tb_writer: SummaryWriter = None
    ):
        if not trainer:
            raise ValueError(f'A {pl.i("trainer")} is required')
        if not hasattr(trainer, 'with_tqdm'):
            raise ValueError(f'Trainer ill-formed: A custom {pl.i("with_tqdm")} field is needed')
        self.trainer = trainer
        self.prettier = prettier
        self.logger = logger
        self.file_logger = file_logger
        self.tb_writer = tb_writer

    def _should_add(self, key: str) -> bool:
        return self.prettier.should_add_split_prefix(key) if self.prettier else True

    def __call__(self, d_log: Dict, training: bool = None, to_console: bool = True):
        training = training or self.trainer.model.training
        d_log_write = self.prettier(d_log) if self.prettier else d_log

        if self.tb_writer:
            tb_step = d_log.get('step') if training else d_log.get('epoch')
            pref = 'train' if training else 'eval'
            for k, v in d_log.items():
                if self._should_add(k):
                    self.tb_writer.add_scalar(tag=f'{pref}/{k}', scalar_value=v, global_step=tb_step)

        if self.trainer.with_tqdm:  # a custom field I added
            pbar = MyProgressCallback.get_current_progress_bar(self.trainer)
            if pbar:
                tqdm_kws = {k: pl.i(v) for k, v in d_log_write.items() if self._should_add(k)}
                pbar.set_postfix(tqdm_kws)
        if to_console:
            if self.logger:
                self.logger.info(pl.i(d_log))

        if self.file_logger:
            self.file_logger.info(pl.nc(d_log))


class CheckArg:
    """
    An easy, readable interface for checking string arguments as effectively enums

    Intended for high-level arguments instead of actual data processing as not as efficient

    Raise errors when common arguments don't match the expected values
    """
    logger = get_logger('Arg Checker')

    def check_mismatch(self, display_name: str, val: str, accepted_values: List[str]):
        if self.verbose:
            d_log = dict(val=val, accepted_values=accepted_values)
            CheckArg.logger.info(f'Checking {pl.i(display_name)} w/ {pl.i(d_log)}... ')
        if val not in accepted_values:
            raise ValueError(f'Unexpected {pl.i(display_name)}: '
                             f'expect one of {pl.i(accepted_values)}, got {pl.i(val)}')

    def __init__(self, ignore_none: bool = True, verbose: bool = False):
        """
        :param ignore_none: If true, arguments passed in as `None` will not raise error
        :param verbose: If true, logging are print to console
        """
        self.d_name2func = dict()
        self.ignore_none = ignore_none
        self.verbose = verbose

    def __call__(self, **kwargs):
        for k, v in kwargs.items():
            if self.ignore_none and v is None:
                if self.verbose:
                    CheckArg.logger.warning(f'Argument {pl.i(k)} is {pl.i("None")} and ignored')
                continue
            self.d_name2func[k](v)

    def cache_mismatch(self, display_name: str, attr_name: str, accepted_values: List[str]):
        self.d_name2func[attr_name] = lambda x: self.check_mismatch(display_name, x, accepted_values)


ca = CheckArg()
ca.cache_mismatch(  # See `stefutil::plot.py`
    'Bar Plot Orientation', attr_name='bar_orient', accepted_values=['v', 'h', 'vertical', 'horizontal']
)


def now(as_str=True, for_path=False, fmt: str = 'full') -> Union[datetime.datetime, str]:
    """
    # Considering file output path
    :param as_str: If true, returns string; otherwise, returns datetime object
    :param for_path: If true, the string returned is formatted as intended for file system path
        relevant only when as_str is True
    :param fmt: One of [`full`, `date`, `short-date`]
        relevant only when as_str is True
    """
    d = datetime.datetime.now()
    if as_str:
        ca.check_mismatch('Date Format', fmt, ['full', 'date', 'short-date'])
        if fmt == 'full':
            fmt_tm = '%Y-%m-%d_%H-%M-%S' if for_path else '%Y-%m-%d %H:%M:%S.%f'
        else:
            fmt_tm = '%Y-%m-%d'
        ret = d.strftime(fmt_tm)
        if fmt == 'short-date':  # year in 2-digits
            ret = ret[2:]
        return ret
    else:
        return d


if __name__ == '__main__':
    # lg = get_logger('test')
    # lg.info('test')

    def check_log_lst():
        lst = ['sda', 'asd']
        print(pl.i(lst))
        # with open('test-logi.txt', 'w') as f:
        #     f.write(pl.nc(lst))
    # check_log_lst()

    def check_log_tup():
        tup = ('sda', 'asd')
        print(pl.i(tup))
    # check_log_tup()

    def check_logi():
        d = dict(a=1, b=2)
        print(pl.i(d))
    # check_logi()

    def check_nested_log_dict():
        d = dict(a=1, b=2, c=dict(d=3, e=4, f=['as', 'as']))
        mic(d)
        print(pl.i(d))
        print(pl.nc(d))
        mic(pl.i(d), pl.nc(d))
    # check_nested_log_dict()

    def check_logger():
        logger = get_logger('blah')
        logger.info('should appear once')
    # check_logger()

    def check_now():
        mic(now(fmt='full'))
        mic(now(fmt='date'))
        mic(now(fmt='short-date'))
    # check_now()

    def check_ca():
        ori = 'v'
        ca(bar_orient=ori)
    # check_ca()

    def check_ca_warn():
        ca_ = CheckArg(verbose=True)
        ca_.cache_mismatch(display_name='Disp Test', attr_name='test', accepted_values=['a', 'b'])
        ca_(test='a')
        ca_(test=None)
    # check_ca_warn()

    def check_time_delta():
        import datetime
        now_ = datetime.datetime.now()
        last_day = now_ - datetime.timedelta(days=1, hours=1, minutes=1, seconds=1)
        mic(now_, last_day)
        diff = now_ - last_day
        mic(diff, fmt_delta(diff))
    check_time_delta()



from .built_in import *
from .os_n_file import *
from .primitive import *
from .container import *
from .prettier import *
from .concurrency import *
from .function import *
from .plot import *
from .ml import *
from .project import *



from setuptools import setup, find_packages

VERSION = '0.17.3'
DESCRIPTION = 'Machine Learning project startup utilities'
LONG_DESCRIPTION = 'My commonly used utilities for machine learning projects'

setup(
    name='stefutils',
    version=VERSION,
    license='MIT',
    author='Yuzhao Stefan Heng',
    author_email='stefan.hg@outlook.com',
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    url='https://github.com/StefanHeng/stef-util',
    download_url='https://github.com/StefanHeng/stef-util/archive/refs/tags/v0.17.3.tar.gz',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'sty', 'colorama', 'pygments',
        'numpy', 'pandas', 'torch', 'transformers',
        'matplotlib', 'seaborn',
        'tqdm', 'icecream'
    ],
    keywords=['python', 'nlp', 'machine-learning', 'deep-learning'],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: MacOS X',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'Operating System :: MacOS',
        'Programming Language :: Python :: 3',
        'Topic :: Utilities'
    ]
)



"""
primitive manipulation
"""

import re
import math
from typing import List, Any, Union

__all__ = ['nan', 'is_float', 'float_is_int', 'clean_whitespace', 'get_substr_indices']


nan = float('nan')


def is_float(x: Any, no_int=False, no_sci=False) -> bool:
    try:
        is_sci = isinstance(x, str) and 'e' in x.lower()
        f = float(x)
        is_int = f.is_integer()
        out = True
        if no_int:
            out = out and (not is_int)
        if no_sci:
            out = out and (not is_sci)
        return out
    except (ValueError, TypeError):
        return False


def float_is_int(f: float, eps: float = None) -> Union[int, bool]:
    if eps:
        return f.is_integer() or math.isclose(f, round(f), abs_tol=eps)
    else:
        return f.is_integer()


def clean_whitespace(s: str):
    if not hasattr(clean_whitespace, 'pattern_space'):
        clean_whitespace.pattern_space = re.compile(r'\s+')
    return clean_whitespace.pattern_space.sub(' ', s).strip()


def get_substr_indices(s: str, s_sub: str) -> List[int]:
    s_sub = re.escape(s_sub)
    return [m.start() for m in re.finditer(s_sub, s)]
